/*
Fast contents-verification utility.
Copyright (C) 2007 One Laptop Per Child Association, Inc.
Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
Written by C. Scott Ananian <cscott@laptop.org>
*/
#define _LARGEFILE64_SOURCE
#define _GNU_SOURCE
#include <dirent.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include "tomcrypt.h"

/* Fast contents manifest verification tool.
 *
 * Speeds on XO B4, build 602:
 *  SHA + RMD: 6m 6s (4m30s user)
 *  SHA only:  4m46s (3m11s user)
 *  RMD only:  2m54s (1m20s user)
 *  no hash :  1m37s (   3s user)
 */

hash_state file_md[2], running_md[2];
enum { SHA=0, RMD=1 };
static char *last_file="";

#define CHECK_RMD 1
#define CHECK_SHA 0
#define CHECK_RMD_ON_MOD 1
#define CHECK_SHA_ON_MOD 1
#define CHECK_RMD_ON_DIR 1
#define CHECK_SHA_ON_DIR 1

#define DIE								\
    do {								\
	fprintf(stderr, "Contents manifest failure at line %d\n", __LINE__); \
	fprintf(stderr, "Last file examined: %s\n", last_file);		\
	exit(1);							\
    } while(0)

#define LINK_MAX 1024
#define SHA_SIZE (256/8)
#define RMD_SIZE (160/8)

struct ug_map *eat_name(FILE *f);
int eat_int(FILE *contents);
void eat_until(FILE *contents, char delimiter);
int eof(FILE *contents);

/////////////// inode list ///////////////////
u_int32_t *pristine_inode_list=NULL;
int pristine_inode_list_size;
static int inode_cmp(const void *a, const void *b) {
    const u_int32_t *aa = a, *bb = b;
    return *aa - *bb;
}
int is_pristine(ino_t inode_num) {
    u_int32_t key=inode_num, *found;
    if (!pristine_inode_list) return 0;
    found = bsearch(&key, pristine_inode_list, pristine_inode_list_size,
		    sizeof(key), inode_cmp);
    if (!found) return 0;
    return (*found) == inode_num; /* be paranoid about sizeof ino_t */
}
void open_inode_list(char *filename) {
    struct stat st;
    int fd;
    fd = open(filename, O_RDONLY);
    if (fd<0) DIE;
    if (fstat(fd, &st) != 0) DIE;
    pristine_inode_list_size = st.st_size / sizeof(*pristine_inode_list);
    pristine_inode_list = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE,
			       fd, 0);
}

/////////////// user/group file parsing ///////////////
struct ug_map {
    int id;
    struct ug_map *next; /* initially a linked list */
    char name[0];
} **user_map, **group_map;
int user_map_size, group_map_size;

int ug_cmp(const void *a, const void *b) {
    struct ug_map * const *aa = a, * const *bb = b;
    return (*aa)->id - (*bb)->id;
}
/* fields 1 and 3 have the name and the id, respectively */
struct ug_map** ug_read(char *filename, int *num_entries) {
    struct ug_map **result, *head = NULL;
    int i=0;
    FILE *f = fopen(filename, "r");
    if (!f) DIE;
    do {
	struct ug_map *entry = eat_name(f);
	eat_until(f, ':');
	entry->id = eat_int(f);
	eat_until(f, '\n');
	entry->next = head;
	head = entry;
	i++;
    } while (!eof(f));
    /* allocate properly sized result array */
    result = malloc(sizeof(*result) * i);
    *num_entries = i;
    for (i=0; head; i++) {
	result[i] = head;
	head = head->next;
	result[i]->next = NULL;
    }
    /* sort & check for duplicates */
    qsort(result, *num_entries, sizeof(*result), ug_cmp);
    for (i=1; i<*num_entries; i++)
	if (result[i-1]->id == result[i]->id) DIE;
    /* done */
    return result;
}
char *ug_lookup(struct ug_map **ug_map, int ug_size, int id) {
    struct ug_map key, *keyp = &key, **e;
    key.id = id;
    e = bsearch(&keyp, ug_map, ug_size, sizeof(*ug_map), ug_cmp);
    if (!e) DIE;
    return (*e)->name;
}

/////////////// directory information /////////////////
struct dir_info {
    int ignore_all_but_sha;
    unsigned char sha[SHA_SIZE], rmd[RMD_SIZE];
    int dl, ml;
    struct dir_info *next;
    char name[0];
};
struct dir_info *create_dir_info(const char *name) {
    struct dir_info *result;
    int len = strlen(name);
    result = malloc(sizeof(*result)+len+1);
    strncpy(result->name, name, len);
    result->name[len] = '\0';
    result->next = NULL;
    result->ignore_all_but_sha = 0;
    return result;
}
struct dir_info *reverse_dir_info(struct dir_info *di) {
    struct dir_info *result = NULL;
    while (di) {
	struct dir_info * di_next, * result_next;
	di_next = di->next;
	result_next = result;

	result = di;
	result->next = result_next;
	
	di = di_next;
    }
    return result;
}

/////////////// parsing functions /////////////////////
char peek(FILE *contents) {
    int c = fgetc(contents);
    if (c==EOF) DIE;
    ungetc(c, contents);
    return c;
}
int eof(FILE *contents) {
    int c = fgetc(contents);
    if (c==EOF) return 1;
    ungetc(c, contents);
    return 0;
}

int eat_ignore_EOF(FILE *contents) {
    int c = fgetc(contents);
    if (c!=EOF) {
	unsigned char cc = c;
	if (CHECK_SHA_ON_DIR) sha256_process(&running_md[SHA], &cc, 1);
	if (CHECK_RMD_ON_DIR) rmd160_process(&running_md[RMD], &cc, 1);
    }
    return c;
}
char eat(FILE *contents) {
    int c = eat_ignore_EOF(contents);
    if (c==EOF) DIE;
    return (char) c;
}

/* gobbles up to and including the delimiter or EOF */
void eat_until(FILE *contents, char delimiter) {
    int c;
    do {
	c = eat_ignore_EOF(contents);
    } while (((char)c) != delimiter && c != EOF);
}

void eat_char(FILE *contents, char expected) {
    char actual = eat(contents);
    if (actual != expected) DIE;
}
void eat_str(FILE *contents, const char *str) {
    while (*str)
	eat_char(contents, *str++);
}

void eat_escaped_char(FILE *contents, char expected) {
    if (expected=='"' || expected=='\\')
	eat_char(contents, '\\');
    eat_char(contents, expected);
}
void eat_escaped_str(FILE *contents, const char *str) {
    eat_char(contents, '"');
    while (*str)
	eat_escaped_char(contents, *str++);
    eat_char(contents, '"');
}

unsigned char hex2int(char c) {
    if (c >= '0' && c <= '9') return c-'0';
    if (c >= 'a' && c <= 'f') return c-'a'+10;
    DIE;
}
unsigned char eat_hexdigit(FILE *contents) {
    return hex2int(eat(contents));
}
unsigned char eat_hexpair(FILE *contents) {
    unsigned char v = eat_hexdigit(contents);
    v = eat_hexdigit(contents) | (v << 4);
    return v;
}

unsigned char eat_digit(FILE *contents) {
    char c = eat(contents);
    if (c >= '0' && c <= '9') return c-'0';
    DIE;
}

/* eat non-zero unsigned integer.  keep result negative because
 * 2's complement value range is one larger for negative numbers. */
int eat_int_neg(FILE *contents) {
    long l=0;
    char c = eat(contents);
    if (c<'1' || c>'9') DIE; /* no leading zeros */
    while(1) {
	l = (l*10) - (c-'0');
	if (l < INT_MIN) DIE; /* overflow */
	c = peek(contents);
	if (c<'0' || c>'9') break;
	c = eat(contents);
    }
    return (int) l;
}

int eat_int(FILE *contents) {
    int result;
    int c = peek(contents);
    if (c=='-') {
	eat_char(contents, '-');
	return eat_int_neg(contents);
    }
    if (c=='0') {
	eat_char(contents, '0');
	return 0;
    }
    result = eat_int_neg(contents);
    if (result == INT_MIN) DIE; /* overflow */
    return -result;
}

struct ug_map *eat_name(FILE *f) {
    struct ug_map *m;
    char buf[NAME_MAX+1], *cp;
    for (cp=buf; cp < buf+sizeof(buf); cp++) {
	*cp = eat(f);
	if (*cp == ':') {
	    *cp = '\0';
	    m = malloc(sizeof(*m)+strlen(buf)+1);
	    strcpy(m->name, buf);
	    return m;
	}
    }
    DIE;
}

void fv_hash_file(const char *filename, int check_sha, int check_rmd) {
    unsigned char buf[16384];
    ssize_t sz;
    int fd;
    fd = open(filename, O_RDONLY | O_LARGEFILE | O_NOATIME | O_NOFOLLOW);
    if (fd <= 0) DIE;
    if (check_sha) sha256_init(&file_md[SHA]);
    if (check_rmd) rmd160_init(&file_md[RMD]);
    while (1) {
	sz = read(fd, buf, sizeof(buf));
	if (sz <= 0) break;
	if (check_sha) sha256_process(&file_md[SHA], buf, sz);
	if (check_rmd) rmd160_process(&file_md[RMD], buf, sz);
    }
    close(fd);
}

void eat_file_hash(FILE *contents, int digest_alg, int digest_size) {
    unsigned char actual[digest_size];
    int i;
    if (digest_alg == SHA)
	sha256_done(&file_md[SHA], actual);
    else if (digest_alg == RMD)
	rmd160_done(&file_md[RMD], actual);
    else DIE;
    /* verify character by character */
    for (i=0; i<digest_size; i++) {
	unsigned char expected = eat_hexpair(contents);
	if (expected != actual[i])
	    DIE;
    }
}

void eat_hashlist(FILE *contents, int check_sha, int check_rmd) {
    eat_str(contents, "[\"");
    if (check_sha)
	eat_file_hash(contents, SHA, SHA_SIZE);
    else {
	int i;
	for (i=0; i<SHA_SIZE; i++)
	    eat_hexpair(contents);
    }
    eat_str(contents, "\",\"");
    if (check_rmd)
	eat_file_hash(contents, RMD, RMD_SIZE);
    else {
	int i;
	for (i=0; i<RMD_SIZE; i++)
	    eat_hexpair(contents);
    }
    eat_str(contents, "\"]");
}

/* returns true if this entry was for a directory */
struct dir_info *
eat_directory_entry(FILE *contents, const struct dirent *d) {
    int seen_d=0, seen_dl=0, seen_g=0, seen_gn=0, seen_h=0;
    int seen_l=0, seen_m=0, seen_ml=0, seen_u=0, seen_un=0;
    int need_d=0, need_dl=0, need_g=0, need_gn=0, need_h=0;
    int need_l=0, need_m=1, need_ml=0, need_u=0, need_un=0;
    struct dir_info *result = NULL;
    struct stat st;
    char c;
    eat_escaped_str(contents, d->d_name);
    eat_str(contents, ":{\""); // must be at least 'm' property

    // get lstat
    if (lstat(d->d_name, &st) != 0) DIE;
    if (S_ISREG(st.st_mode)) { need_h=1; }
    if (S_ISLNK(st.st_mode)) { need_l=1; }
    if (S_ISDIR(st.st_mode)) { need_h=need_dl=need_ml=1;
	                       result=create_dir_info(d->d_name); }
    if (S_ISCHR(st.st_mode)) { need_d=1; }
    if (S_ISBLK(st.st_mode)) { need_d=1; }

    // parse properties in order: d,dl,g,g#,h,l,m,ml,u,u#
    c = eat(contents);
#define EAT_COLON(attr, suffix)		 \
    do {				 \
	seen_ ## attr = 1;		 \
	eat_str(contents, suffix "\":"); \
    } while(0)
#define NEXT				    \
    do {				    \
	c = eat(contents);		    \
	if (c!=',') goto done;		    \
	eat_char(contents, '"');	    \
	c = eat(contents);		    \
    } while(0)
    
    if (c=='d' && peek(contents)=='"') {
	EAT_COLON(d, "");
	if (st.st_rdev != eat_int(contents)) DIE;
	NEXT;
    }
    if (c=='d' && peek(contents)=='l') {
	EAT_COLON(dl, "l");
	result->dl = eat_int(contents);
	NEXT;
    }
    if (c=='g' && peek(contents)=='"') {
	EAT_COLON(g, "");
	eat_escaped_str(contents,
			ug_lookup(group_map, group_map_size, st.st_gid));
	NEXT;
    }
    if (c=='g' && peek(contents)=='#') {
	EAT_COLON(gn, "#");
	if (st.st_gid != eat_int(contents)) DIE;
	NEXT;
    }
    if (c=='h') {
	EAT_COLON(h, "");
	if (S_ISREG(st.st_mode)) {
	    int pristine = is_pristine(st.st_ino);
	    int check_sha = pristine ? CHECK_SHA : CHECK_SHA_ON_MOD;
	    int check_rmd = pristine ? CHECK_RMD : CHECK_RMD_ON_MOD;
	    fv_hash_file(d->d_name, check_sha, check_rmd);
	    eat_hashlist(contents, check_sha, check_rmd);
	} else if (S_ISDIR(st.st_mode)) {
	    int i;
	    eat_str(contents, "[\"");
	    for (i=0; i<SHA_SIZE; i++)
		result->sha[i] = eat_hexpair(contents);
	    eat_str(contents, "\",\"");
	    for (i=0; i<RMD_SIZE; i++)
		result->rmd[i] = eat_hexpair(contents);
	    eat_str(contents, "\"]");
	} else DIE;
	NEXT;
    }
    if (c=='l') {
	char buf[LINK_MAX];
	ssize_t sz;
	EAT_COLON(l, "");
	if (!S_ISLNK(st.st_mode)) DIE;
	sz = readlink(d->d_name, buf, sizeof(buf));
	if (sz<0 || sz==sizeof(buf)) DIE;
	buf[sz]='\0'; /* readlink doesn't null-terminate its buffer =( */
	eat_escaped_str(contents, buf);
	NEXT;
    }
    if (c=='m' && peek(contents)=='"') {
	EAT_COLON(m, "");
	if (st.st_mode != eat_int(contents)) DIE;
	NEXT;
    }
    if (c=='m' && peek(contents)=='l') {
	EAT_COLON(ml, "l");
	result->ml = eat_int(contents);
	NEXT;
    }
    if (c=='u' && peek(contents)=='"') {
	EAT_COLON(u, "");
	eat_escaped_str(contents,
			ug_lookup(user_map, user_map_size, st.st_uid));
	NEXT;
    }
    if (c=='u' && peek(contents)=='#') {
	EAT_COLON(un, "#");
	if (st.st_uid != eat_int(contents)) DIE;
	NEXT;
    }
    DIE; /* some weird property */
 done:
    if (c!='}') DIE;

    if (seen_d < need_d || seen_dl< need_dl|| seen_g < need_g ||
	seen_gn< need_gn|| seen_h < need_h || seen_l < need_l ||
	seen_m < need_m || seen_ml< need_ml|| seen_u < need_u ||
	seen_un< need_un)
	DIE;

    /* okay! */
    return result;
}

typedef typeof(&alphasort) scandir_sort_t;
static int dirsort(const struct dirent ** a, const struct dirent ** b) {
    // lexicographic sort.
    return strcmp((*a)->d_name, (*b)->d_name);
}
static int dirfilter(const struct dirent * a) {
    const char *name = a->d_name;
    if (*name == '.') {
	if (strcmp(name, ".") == 0 ||
	    strcmp(name, "..") == 0 ||
	    strcmp(name, ".xo-files") == 0)
	    return 0;
    }
    return 1;
}

void eat_directory_object(FILE *contents, struct dir_info *di_expected) {
    unsigned char sha[SHA_SIZE], rmd[RMD_SIZE];
    struct dir_info *dirs = NULL;
    struct dirent **namelist;
    char *prev_file = last_file;
    long start_pos;
    int n, i=0;
    // get directory contents.
    n = scandir(".", &namelist, dirfilter, (scandir_sort_t) dirsort);
    // parse contents
    start_pos = ftell(contents);
    if (CHECK_SHA_ON_DIR) sha256_init(&running_md[SHA]);
    if (CHECK_RMD_ON_DIR) rmd160_init(&running_md[RMD]);

    eat_str(contents, "[\"dir\",1,[[\"sha-256\",\"ripemd-160\"],{");
    // no items in list?
    if (peek(contents)!='}')
	while(1) {
	    struct dir_info *d;
	    if (i>=n) DIE;
	    // do another entry.
	    last_file=namelist[i]->d_name; // for better failure diagnostics
	    d = eat_directory_entry(contents, namelist[i]);
	    if (d) { // add directory to our list.
		d->next = dirs;
		dirs = d;
	    }
	    last_file=prev_file; // reset
	    free(namelist[i++]);
	    if (peek(contents)!=',') break;
	    eat_char(contents, ',');
	}
    eat_char(contents, '}');
    if (i!=n) DIE;
    eat_str(contents, "]]");

    /* check dl */
    if (di_expected->dl != (ftell(contents)-start_pos) &&
	!di_expected->ignore_all_but_sha) DIE;
    /* check dir hash */
    if (CHECK_SHA_ON_DIR) {
	sha256_done(&running_md[SHA], sha);
	if (memcmp(di_expected->sha, sha, SHA_SIZE) != 0)
	    DIE; /* directory hash mismatch */
    }
    if (CHECK_RMD_ON_DIR &&
	!di_expected->ignore_all_but_sha) {
	rmd160_done(&running_md[RMD], rmd);
	if (memcmp(di_expected->rmd, rmd, RMD_SIZE) != 0)
	    DIE; /* directory hash mismatch */
    }

    // eat children.
    dirs = reverse_dir_info(dirs);
    while (dirs) {
	struct dir_info *d_next = dirs->next;
	eat_char(contents, ',');
	last_file=dirs->name; // for better failure diagnostics
	if (chdir(dirs->name) != 0) DIE;
	eat_directory_object(contents, dirs);
	if (chdir("..") != 0) DIE;
	last_file=prev_file;  // reset
	free(dirs);
	dirs = d_next;
    }

    if ((di_expected->ml - 17) != (ftell(contents)-start_pos) &&
	!di_expected->ignore_all_but_sha)
	DIE; /* bad ml */
}

void eat_contents(FILE *contents, char *dir, char *expected_sha_hash) {
    struct dir_info di;
    char *prev_file = last_file;
    int i;
    last_file=dir; // for better failure diagnostics
    if (chdir(dir) != 0) DIE;
    eat_str(contents, "[\"manifest\",1,[");
    di.ignore_all_but_sha = 1;
    if (strlen(expected_sha_hash)!=SHA_SIZE*2) DIE;
    for (i=0; i<SHA_SIZE; i++)
	di.sha[i] = (hex2int(expected_sha_hash[2*i]) << 4) |
	    hex2int(expected_sha_hash[2*i + 1]);
    eat_directory_object(contents, &di);
    eat_str(contents, "]]");
    if (!eof(contents)) DIE;
    last_file=prev_file; // dir may go away when we return.
    /* success! */
}

int main(int argc, char **argv) {
    FILE *contents;

    if (argc != 6 && argc != 7) {
	fprintf(stderr, "Usage: %s <passwd file> <group file> "
		"<contents hash> "
		"<contents object> <directory name> [<pristine inodes>]\n",
		argv[0]);
	exit(3);
    }

    user_map = ug_read(argv[1], &user_map_size);
    group_map = ug_read(argv[2], &group_map_size);

    if (argc==7) open_inode_list(argv[6]);

    contents = fopen(argv[4], "rb");
    if (!contents) {
	perror("Contents manifest file not found");
	exit(3);
    }
    eat_contents(contents, argv[5], argv[3]);
    fclose(contents);

    return 0; /* success! */
}
